import typer
from pathlib import Path

from typerdrive.cache.exceptions import CacheError
from typerdrive.cache.attach import get_manager, attach_cache
from typerdrive.cache.manager import CacheManager

from tests.helpers import check_output, match_output


class TestAttachCache:
    def test_attach_cache__adds_cache_manager_to_context(self):
        cli: typer.Typer = typer.Typer()

        @cli.command()
        @attach_cache()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = ctx.obj.cache_manager
            assert isinstance(manager, CacheManager)
            assert manager.app_name == "test"

        check_output(cli, prog_name="test")

    def test_attach_cache__show(self, fake_cache_path: Path):
        cli = typer.Typer()

        @cli.command()
        @attach_cache(show=True)
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            manager.store_bytes(b"jawa", "jawa")
            manager.store_bytes(b"ewok", "ewok")
            manager.store_bytes(b"hutt & pyke", "hutt/pyke")

        expected_pattern = f"""
            📂 {fake_cache_path}
            ├── 📂 hutt
            │   └── 📄pyke (11 Bytes)
            ├── 📄ewok (4 Bytes)
            └── 📄jawa (4 Bytes)
        """
        match_output(
            cli,
            expected_pattern=expected_pattern,
            escape_parens=True,
            prog_name="test",
        )


class TestWithParameters:
    def test_attach_cache__with_manager_parameter(self, fake_cache_path: Path):
        cli = typer.Typer()

        @cli.command()
        @attach_cache(show=True)
        def noop(ctx: typer.Context, mgr: CacheManager):  # pyright: ignore[reportUnusedFunction, reportUnusedParameter]
            mgr.store_bytes(b"jawa", "jawa")
            mgr.store_bytes(b"ewok", "ewok")
            mgr.store_bytes(b"hutt & pyke", "hutt/pyke")

        expected_pattern = f"""
            📂 {fake_cache_path}
            ├── 📂 hutt
            │   └── 📄pyke (11 Bytes)
            ├── 📄ewok (4 Bytes)
            └── 📄jawa (4 Bytes)
        """
        match_output(
            cli,
            expected_pattern=expected_pattern,
            escape_parens=True,
            prog_name="test",
        )


class TestGetManager:
    def test_get_manager__extracts_cache_manager_from_context(self, fake_cache_path: Path):
        cli = typer.Typer()

        @cli.command()
        @attach_cache()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            manager = get_manager(ctx)
            assert isinstance(manager, CacheManager)
            assert manager.app_name == "test"
            assert manager.cache_path == fake_cache_path
            print("Passed!")

        match_output(
            cli,
            expected_pattern=["Passed"],
            exit_code=0,
            prog_name="test",
        )

    def test_get_manager__raises_exception_if_context_has_no_manager(self):
        cli = typer.Typer()

        @cli.command()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            get_manager(ctx)
            print("Passed!")

        match_output(
            cli,
            exception_type=CacheError,
            exception_pattern="Cache is not bound to the context",
            exit_code=1,
            prog_name="test",
        )

    def test_get_manager__raises_exception_if_non_manager_retrieved(self):
        cli = typer.Typer()

        @cli.command()
        @attach_cache()
        def noop(ctx: typer.Context):  # pyright: ignore[reportUnusedFunction]
            ctx.obj.cache_manager = "Not a manager!"
            get_manager(ctx)
            print("Passed!")

        match_output(
            cli,
            exception_type=CacheError,
            exception_pattern="Item in user context at `cache_manager` was not a CacheManager",
            exit_code=1,
            prog_name="test",
        )



from typing import Annotated

from pydantic import AfterValidator, BaseModel


def valid_alignment(value: str) -> str:
    if value not in ["good", "neutral", "evil"]:
        raise ValueError(f"{value} is an invalid alignment")
    return value


class DefaultSettingsModel(BaseModel):
    name: str = "jawa"
    planet: str = "tatooine"
    is_humanoid: bool = True
    alignment: Annotated[str, AfterValidator(valid_alignment)] = "neutral"


class RequiredFieldsModel(BaseModel):
    name: str
    planet: str
    is_humanoid: bool = True
    alignment: Annotated[str, AfterValidator(valid_alignment)] = "neutral"



from importlib.metadata import PackageNotFoundError
from unittest import mock

from typerdrive.version import (
    __version__,
    get_version,
    get_version_from_metadata,
    get_version_from_pyproject,
)


def test_get_version_from_metadata():
    assert get_version_from_metadata() == __version__


def test_get_version_from_pyproject():
    assert get_version_from_pyproject() == __version__


def test_get_version__uses_pyproject():
    expected_version = "1.2.3"
    with mock.patch(
        "typerdrive.version.get_version_from_metadata", side_effect=PackageNotFoundError
    ) as mocked_metadata:
        with mock.patch(
            "typerdrive.version.get_version_from_pyproject", return_value=expected_version
        ) as mocked_pyproject:
            computed_version = get_version()
            assert computed_version == expected_version

            mocked_metadata.assert_called_once()
            mocked_pyproject.assert_called_once()


def test_get_version__uses_metadata():
    expected_version = "1.2.3"
    with mock.patch("typerdrive.version.get_version_from_metadata", return_value=expected_version) as mocked_metadata:
        with mock.patch(
            "typerdrive.version.get_version_from_pyproject", side_effect=FileNotFoundError
        ) as mocked_pyproject:
            computed_version = get_version()
            assert computed_version == expected_version

            mocked_metadata.assert_called_once()
            mocked_pyproject.assert_not_called()


def test_get_version__returns_unknown_if_both_fail():
    expected_version = "unknown"
    with mock.patch(
        "typerdrive.version.get_version_from_metadata", side_effect=PackageNotFoundError
    ) as mocked_metadata:
        with mock.patch(
            "typerdrive.version.get_version_from_pyproject", side_effect=FileNotFoundError
        ) as mocked_pyproject:
            computed_version = get_version()
            assert computed_version == expected_version

            mocked_metadata.assert_called_once()
            mocked_pyproject.assert_called_once()



from enum import StrEnum, auto
import traceback
import logging
import random

from buzz import DoExceptParams
import typer

from typerdrive.exceptions import TyperdriveError, handle_errors
from typerdrive.format import terminal_message, strip_rich_style


class CallIt(StrEnum):
    heads = auto()
    tails = auto()


logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
console_handler = logging.StreamHandler()
console_handler.setFormatter(logging.Formatter('%(asctime)s: %(levelname)s -> %(message)s'))
logger.addHandler(console_handler)


def log_error(params: DoExceptParams):
    logger.error(
        "\n".join(
            [
                strip_rich_style(params.final_message),
                "--------",
                "Traceback:",
                "".join(traceback.format_tb(params.trace)),
            ]
        )
    )


def log_success():
    logger.info("No errors occurred!")


def log_done():
    logger.info("Program complete. Exiting.")


cli = typer.Typer()


@cli.command()
@handle_errors(
    "Flip error",
    do_except=log_error,
    do_else=log_success,
    do_finally=log_done,
)
def flip(call_it: CallIt, show_logs: bool = False):
    if show_logs:
        logger.setLevel(logging.DEBUG)
    result = random.choice([c for c in CallIt])
    logger.debug(f"Result: {result}")
    if call_it != result:
        raise TyperdriveError(
            f"[yellow]{result}[/yellow], [red]you lose![/red]",
            subject="Womp, womp",
            footer="Don't sweat it; just try again!",
        )
    terminal_message(
        f"[yellow]{result}[/yellow], [green]you win![/green]",
        subject="Tada!",
        footer="Maybe you won't be so lucky next time!",
    )


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
import json
from pathlib import Path
from typing import Any

import humanize
from rich.markup import escape
from rich.text import Text
from rich.tree import Tree

from typerdrive.cache.exceptions import (
    CacheClearError,
    CacheError,
    CacheInitError,
    CacheLoadError,
    CacheStoreError,
)


def get_cache_path(app_name: str) -> Path:
    return Path.home() / ".cache" / app_name


def _clear_dir(path: Path) -> int:
    count = 0
    for sub_path in path.iterdir():
        if sub_path.is_dir():
            count += _clear_dir(sub_path)
            sub_path.rmdir()
        else:
            sub_path.unlink()
            count += 1
    return count


def clear_directory(path: Path) -> int:
    CacheError.require_condition(path.exists(), f"Target {path=} does not exist")
    CacheError.require_condition(path.is_dir(), f"Target {path=} is not a directory")
    with CacheError.handle_errors(f"Failed to clear directory at {path=}"):
        return _clear_dir(path)


@dataclass
class CacheInfo:
    tree: Tree
    file_count: int
    total_size: int


def render_directory(path: Path, is_root: bool = True) -> CacheInfo:
    root_label: str
    if is_root:
        root_label = str(path)
        color = "[bold yellow]"
    else:
        root_label = escape(path.name)
        color = "[bold blue]"

    cache_info = CacheInfo(
        tree=Tree(f"{color}📂 {root_label}"),
        file_count=0,
        total_size=0,
    )

    child_paths = sorted(
        path.iterdir(),
        key=lambda p: (
            p.is_file(),
            p.name.lower(),
        ),
    )
    for child_path in child_paths:
        if child_path.is_dir():
            child_info = render_directory(child_path, is_root=False)
            cache_info.tree
            cache_info.tree.children.append(child_info.tree)
            cache_info.file_count += child_info.file_count
            cache_info.total_size += child_info.total_size
        else:
            file_size = child_path.stat().st_size
            icon = Text("📄 ")
            label = Text(escape(child_path.name), "green")
            info = Text(f" ({humanize.naturalsize(file_size)})", "blue")
            cache_info.tree.add(icon + label + info)
            cache_info.file_count += 1
            cache_info.total_size += file_size
    return cache_info


def is_child(path: Path, parent: Path):
    root_path = Path(path.parts[0])
    temp_path = path
    while temp_path != root_path:
        if temp_path == parent:
            return True
        temp_path = temp_path.parent
    return False


# TODO: add a mechanism to cleanup empty directories
class CacheManager:
    def __init__(self, app_name: str):
        self.app_name: str = app_name
        self.cache_path: Path = get_cache_path(self.app_name)
        with CacheInitError.handle_errors("Failed to initialize cache"):
            self.cache_path.mkdir(parents=True, exist_ok=True)

    def resolve_path(self, path: Path | str, mkdir: bool = False) -> Path:
        if isinstance(path, str):
            path = Path(path)
        full_path = self.cache_path / path
        full_path = full_path.resolve()
        CacheError.require_condition(
            is_child(full_path, self.cache_path),
            f"Resolved cache path {str(full_path)} is not within cache {str(self.cache_path)}",
        )
        CacheError.require_condition(
            full_path != self.cache_path,
            f"Resolved cache path {str(full_path)} must not be the same as cache {str(self.cache_path)}",
        )
        if mkdir:
            full_path.parent.mkdir(parents=True, exist_ok=True)
        return full_path

    def store_bytes(self, data: bytes, path: Path | str, mode: int | None = None):
        full_path = self.resolve_path(path, mkdir=True)
        with CacheStoreError.handle_errors(f"Failed to store data in cache target {str(path)}"):
            full_path.write_bytes(data)
        if mode:
            with CacheStoreError.handle_errors(f"Failed to set mode for cache target {str(path)} to {mode=}"):
                full_path.chmod(mode)

    def store_text(self, text: str, path: Path | str, mode: int | None = None):
        self.store_bytes(text.encode("utf-8"), path, mode=mode)

    def store_json(self, data: dict[str, Any], path: Path | str, mode: int | None = None):
        self.store_bytes(json.dumps(data, indent=2).encode("utf-8"), path, mode=mode)

    def load_bytes(self, path: Path | str) -> bytes:
        full_path = self.resolve_path(path, mkdir=False)
        CacheLoadError.require_condition(full_path.exists(), f"Cache target {str(path)} does not exist")
        with CacheLoadError.handle_errors(f"Failed to load data from cache target {str(path)}"):
            return full_path.read_bytes()

    def load_text(self, path: Path | str) -> str:
        return self.load_bytes(path).decode("utf-8")

    def load_json(self, path: Path | str) -> dict[str, Any]:
        text = self.load_bytes(path).decode("utf-8")
        with CacheLoadError.handle_errors(f"Failed to unpack JSON data from cache target {str(path)}"):
            return json.loads(text)

    def clear_path(self, path: Path | str) -> Path:
        full_path = self.resolve_path(path)
        with CacheClearError.handle_errors(f"Failed to clear cache target {str(path)}"):
            full_path.unlink()
        if len([p for p in full_path.parent.iterdir()]) == 0:
            with CacheClearError.handle_errors(f"Failed to remove empty directory {str(full_path.parent)}"):
                full_path.parent.rmdir()
        return full_path

    def clear_all(self) -> int:
        with CacheClearError.handle_errors("Failed to clear cache"):
            return clear_directory(self.cache_path)

    def pretty(self) -> CacheInfo:
        return render_directory(self.cache_path)



from itertools import pairwise
import re
from typing import Any

import typer
from snick import conjoin, strip_ansi_escape_sequences, strip_whitespace, dedent, indent
from typer.testing import CliRunner


runner = CliRunner()


def strip_lineart(text: str) -> str:
    lineart = ["╭", "─", "┬", "╮", "│", "├", "┼", "┤", "╰", "┴", "╯", "└"]
    for char in lineart:
        text = text.replace(char, "")
    return text


def get_output(
    cli: typer.Typer,
    *args: str,
    exit_code: int = 0,
    env_vars: dict[str, str] | None = None,
    strip_terminal_controls: bool = True,
    exception_type: type[Exception] | None = None,
    exception_pattern: str | None = None,
    **kwargs: Any,
) -> str:
    if env_vars is None:
        env_vars = {}
    result = runner.invoke(cli, args, env=env_vars, **kwargs)
    output = result.stdout
    if strip_terminal_controls:
        output = strip_ansi_escape_sequences(output)
    assert result.exit_code == exit_code, build_code_fail_message(exit_code, result.exit_code, output, result.exception)
    if exception_type:
        assert isinstance(result.exception, exception_type), build_exception_type_message(
            result.exception,
            exception_type,
            output,
        )
    if exception_pattern:
        assert re.search(exception_pattern, str(result.exception)), build_exception_pattern_message(
            result.exception,
            exception_pattern,
            output,
        )

    return output


def get_help(cli: typer.Typer) -> str:
    return get_output(cli, "--help")


def check_output(
    cli: typer.Typer,
    *args: str,
    expected_substring: str | list[str] | None = None,
    unwanted_substring: str | list[str] | None = None,
    **kwargs: Any,
):
    output = get_output(cli, *args, **kwargs)
    if expected_substring:
        if isinstance(expected_substring, str):
            expected_substring = [expected_substring]
        for es in expected_substring:
            assert es in output, build_substring_fail_message(es, output)
    if unwanted_substring:
        if isinstance(unwanted_substring, str):
            unwanted_substring = [unwanted_substring]
        for us in unwanted_substring:
            assert us not in output, build_substring_fail_message(us, output, negative_match=True)


def check_help(cli: typer.Typer, **kwargs: Any):
    check_output(cli, "--help", exit_code=0, **kwargs)


def match_output(
    cli: typer.Typer,
    *args: str,
    expected_pattern: str | list[str] | None = None,
    unwanted_pattern: str | list[str] | None = None,
    enforce_order: bool = True,
    escape_parens: bool = False,
    escape_brackets: bool = False,
    **kwargs: Any,
):
    output = get_output(cli, *args, **kwargs)
    if expected_pattern:
        if isinstance(expected_pattern, str):
            expected_pattern = [expected_pattern]

        start_positions: list[int] = []
        for ep in expected_pattern:
            mangled_pattern = ep
            mangled_pattern = strip_lineart(strip_whitespace(mangled_pattern))
            if escape_parens:
                mangled_pattern = mangled_pattern.replace("(", r"\(").replace(")", r"\)")
            if escape_brackets:
                mangled_pattern = mangled_pattern.replace("[", r"\[").replace("]", r"\]")

            mangled_output = output
            mangled_output = strip_lineart(strip_whitespace(mangled_output))

            match: re.Match[str] | None = re.search(mangled_pattern, mangled_output)
            if match is not None:
                start_positions.append(match.start())

            assert match, build_pattern_fail_message(ep, mangled_pattern, output, mangled_output)

        if enforce_order:
            assert all(left <= right for (left, right) in pairwise(start_positions)), build_order_fail_message(
                expected_pattern, start_positions, output
            )

    if unwanted_pattern:
        if isinstance(unwanted_pattern, str):
            unwanted_pattern = [unwanted_pattern]

        for up in unwanted_pattern:
            mangled_pattern = up
            mangled_pattern = strip_lineart(strip_whitespace(mangled_pattern))
            if escape_parens:
                mangled_pattern = mangled_pattern.replace("(", r"\(").replace(")", r"\)")
            if escape_brackets:
                mangled_pattern = mangled_pattern.replace("[", r"\[").replace("]", r"\]")

            mangled_output = output
            mangled_output = strip_lineart(strip_whitespace(mangled_output))

            assert not re.search(mangled_pattern, mangled_output), build_pattern_fail_message(
                up,
                mangled_pattern,
                output,
                mangled_output,
                negative_pattern=True,
            )


def match_help(
    cli: typer.Typer,
    **kwargs: Any,
):
    match_output(cli, "--help", exit_code=0, **kwargs)


def build_code_fail_message(
    expected_code: int,
    computed_code: int,
    output: str,
    exception: BaseException | None,
) -> str:
    return dedent(
        f"""
        Exit codes didn't match!

        Expected {expected_code}
        Computed {computed_code}

        Exception:
        {exception}

        Output:
        {indent(output, prefix="            ", skip_first_line=True)}
        """
    )


def build_exception_type_message(
    exception: BaseException | None,
    exception_type: type[Exception],
    output: str,
) -> str:
    return dedent(
        f"""
        Expected exception type doesn't match!

        Expected {exception_type}
        Computed {type[exception]}

        Exception:
        {exception}

        Output:
        {indent(output, prefix="            ", skip_first_line=True)}
        """
    )


def build_exception_pattern_message(
    exception: BaseException | None,
    exception_pattern: str,
    output: str,
) -> str:
    return dedent(
        f"""
        Expected exception text doesn't match pattern!

        Expected {exception_pattern}
        Computed {exception}

        Output:
        {indent(output, prefix="            ", skip_first_line=True)}
        """
    )


def build_substring_fail_message(
    substring: str,
    output: str,
    negative_match: bool = False,
) -> str:
    qualifier = "was not found" if not negative_match else "WAS FOUND"
    return dedent(
        f"""
        Substring {qualifier} in output

        Substring:
        {substring}

        Output:
        {indent(output, prefix="            ", skip_first_line=True)}
        """
    )


def build_pattern_fail_message(
    pattern: str,
    mangled_pattern: str,
    output: str,
    mangled_output: str,
    negative_pattern: bool = False,
) -> str:
    qualifier = "was not found" if not negative_pattern else "WAS FOUND"
    return dedent(
        f"""
        Search pattern {qualifier} in output

        Search Pattern:
        {pattern}

        "Mangled" Search Pattern:
        {repr(mangled_pattern)}

        Output:
        {indent(output, prefix="            ", skip_first_line=True)}

        "Mangled" Output:
        {repr(mangled_output)}
        """
    )


def build_order_fail_message(expected_patterns: list[str], start_positions: list[int], output: str) -> str:
    return conjoin(
        "Search patterns were out of order",
        "",
        *[f"{i}: {p}" for (i, p) in zip(start_positions, expected_patterns)],
        "",
        "Output:",
        output,
    )



import json
from pathlib import Path

import pytest
from pytest_mock import MockerFixture
from snick import dedent

from typerdrive.cache.exceptions import CacheError, CacheInitError
from typerdrive.cache.manager import CacheManager, clear_directory

from tests.cache.helpers import tree_to_text


class TestSettingsManager:
    def test_init__no_issues(self, fake_cache_path: Path):
        manager = CacheManager("test")
        assert manager.app_name == "test"
        assert manager.cache_path == fake_cache_path

    def test_init__raises_exception_on_fail(self, mocker: MockerFixture):
        mocker.patch("typerdrive.cache.manager.Path.mkdir", side_effect=RuntimeError("BOOM!"))
        with pytest.raises(CacheInitError, match="Failed to initialize cache"):
            CacheManager("test")

    def test_resolve_path__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        full_path = manager.resolve_path(test_path)
        assert full_path == fake_cache_path / test_path

    def test_resolve_path__fails_if_path_is_outside_cache(self, fake_cache_path: Path):
        test_path = Path("../jawa")
        manager = CacheManager("test")
        full_path = fake_cache_path / test_path
        with pytest.raises(CacheError, match=f"Resolved cache path .* is not within cache {str(fake_cache_path)}"):
            manager.resolve_path(full_path)

    def test_resolve_path__fails_if_path_is_the_same_as_cache_path(self, fake_cache_path: Path):
        test_path = Path(".")
        manager = CacheManager("test")
        full_path = fake_cache_path / test_path
        with pytest.raises(
            CacheError,
            match=f"Resolved cache path .* must not be the same as cache {str(fake_cache_path)}",
        ):
            manager.resolve_path(full_path)

    def test_resolve_path__makes_parents_if_requested(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        full_path = manager.resolve_path(test_path, mkdir=True)
        assert full_path == fake_cache_path / test_path
        assert full_path.parent.exists()

    def test_resolve_path__works_with_strings(self, fake_cache_path: Path):
        test_path = "jawa/ewok/hutt"
        manager = CacheManager("test")
        full_path = manager.resolve_path(test_path)
        assert full_path == fake_cache_path / test_path

    def test_store_bytes__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = b"test data"
        manager.store_bytes(data, test_path)
        full_path = fake_cache_path / test_path
        assert full_path.read_bytes() == data

    def test_store_bytes__raises_error_if_write_fails(self, fake_cache_path: Path, mocker: MockerFixture):
        mocker.patch("typerdrive.cache.manager.Path.write_bytes", side_effect=RuntimeError("BOOM!"))
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = b"test data"
        with pytest.raises(CacheError, match="Failed to store data in cache target jawa/ewok/hutt"):
            manager.store_bytes(data, test_path)
        full_path = fake_cache_path / test_path
        assert not full_path.exists()

    def test_store_bytes__sets_mode(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = b"test data"
        mode = 0o141
        manager.store_bytes(data, test_path, mode=mode)
        full_path = fake_cache_path / test_path
        assert full_path.stat().st_mode & 0o777 == mode

    def test_store_bytes__raises_error_if_chmod_fails(self, fake_cache_path: Path, mocker: MockerFixture):
        mocker.patch("typerdrive.cache.manager.Path.chmod", side_effect=RuntimeError("BOOM!"))
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = b"test data"
        mode = 0o141
        with pytest.raises(CacheError, match=f"Failed to set mode for cache target jawa/ewok/hutt to {mode=}"):
            manager.store_bytes(data, test_path, mode=mode)
        full_path = fake_cache_path / test_path
        assert full_path.stat().st_mode & 0o777 != mode

    def test_store_text__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = "test data"
        manager.store_text(data, test_path)
        full_path = fake_cache_path / test_path
        assert full_path.read_text() == data

    def test_store_json__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = dict(name="sleazy", species="hutt")
        manager.store_json(data, test_path)
        full_path = fake_cache_path / test_path
        assert json.loads(full_path.read_text()) == data

    def test_load_bytes__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = b"test data"
        full_path = fake_cache_path / test_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_bytes(data)
        loaded_data = manager.load_bytes(test_path)
        assert loaded_data == data

    def test_load_bytes__raises_error_if_file_does_not_exist(self):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        with pytest.raises(CacheError, match="Cache target jawa/ewok/hutt does not exist"):
            manager.load_bytes(test_path)

    def test_load_bytes__fails_if_read_fails(self, fake_cache_path: Path, mocker: MockerFixture):
        mocker.patch("typerdrive.cache.manager.Path.read_bytes", side_effect=RuntimeError("BOOM!"))
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = b"test data"
        full_path = fake_cache_path / test_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_bytes(data)
        with pytest.raises(CacheError, match="Failed to load data from cache target jawa/ewok/hutt"):
            manager.load_bytes(test_path)

    def test_load_text__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = "test data"
        full_path = fake_cache_path / test_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_bytes(data.encode("utf-8"))
        loaded_data = manager.load_text(test_path)
        assert loaded_data == data

    def test_load_json__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        data = dict(name="sleazy", species="hutt")
        full_path = fake_cache_path / test_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_bytes(json.dumps(data).encode("utf-8"))
        loaded_data = manager.load_json(test_path)
        assert loaded_data == data

    def test_clear_path__basic(self, fake_cache_path: Path):
        test_path = Path("jawa/ewok/hutt")
        other_path = Path("jawa/ewok/pyke")
        manager = CacheManager("test")
        data = b"test data"
        other_data = b"other data"
        manager.store_bytes(data, test_path)
        manager.store_bytes(other_data, other_path)
        full_path = manager.clear_path(test_path)
        assert full_path == fake_cache_path / test_path
        assert not full_path.exists()

        other_full_path = fake_cache_path / other_path
        assert other_full_path.read_bytes() == other_data

    def test_clear_path__raises_error_if_unlink_fails(self, mocker: MockerFixture):
        mocker.patch("typerdrive.cache.manager.Path.unlink", side_effect=RuntimeError("BOOM!"))
        test_path = Path("jawa/ewok/hutt")
        manager = CacheManager("test")
        with pytest.raises(CacheError, match="Failed to clear cache target jawa/ewok/hutt"):
            manager.clear_path(test_path)

    def test_clear_all__basic(self, fake_cache_path: Path):
        manager = CacheManager("test")
        manager.store_bytes(b"jawa", "jawa")
        manager.store_bytes(b"ewok", "ewok")
        manager.store_bytes(b"hutt & pyke", "hutt/pyke")
        assert len([p for p in fake_cache_path.iterdir()]) == 3
        manager.clear_all()
        assert len([p for p in fake_cache_path.iterdir()]) == 0

    def test_clear_all__raises_error_if_clear_directory_fails(self, mocker: MockerFixture):
        mocker.patch("typerdrive.cache.manager.clear_directory", side_effect=RuntimeError("BOOM!"))
        manager = CacheManager("test")
        with pytest.raises(CacheError, match="Failed to clear cache"):
            manager.clear_all()

    def test_pretty(self, fake_cache_path: Path):
        manager = CacheManager("test")
        manager.store_bytes(b"jawa", "jawa")
        manager.store_bytes(b"ewok", "ewok")
        manager.store_bytes(b"hutt & pyke", "hutt/pyke")
        cache_info = manager.pretty()
        expected = dedent(f"""
            📂 {fake_cache_path}
            ├── 📂 hutt
            │   └── 📄 pyke (11 Bytes)
            ├── 📄 ewok (4 Bytes)
            └── 📄 jawa (4 Bytes)
        """)
        assert tree_to_text(cache_info.tree) == expected


class TestHelpers:
    def test_clear_directory__basic(self, tmp_path: Path):
        tmp_path.joinpath("jawa/ewok").mkdir(parents=True)
        tmp_path.joinpath("jawa/ewok/talz").touch()

        tmp_path.joinpath("pyke/hutt").mkdir(parents=True)
        tmp_path.joinpath("pyke/hutt/muun").touch()

        tmp_path.joinpath("bith").mkdir(parents=True)
        tmp_path.joinpath("bith/gran").touch()

        tmp_path.joinpath("teek").touch()

        assert len([p for p in tmp_path.iterdir()]) == 4
        clear_directory(tmp_path)
        assert len([p for p in tmp_path.iterdir()]) == 0

    def test_clear_directory__raises_error_if_path_does_not_exist(self):
        with pytest.raises(CacheError, match="Target path=.* does not exist"):
            clear_directory(Path("nonexistent"))

    def test_clear_directory__raises_error_if_path_is_not_a_directory(self, tmp_path: Path):
        nondir = tmp_path / "nondir"
        nondir.touch()
        with pytest.raises(CacheError, match="Target path=.* is not a directory"):
            clear_directory(nondir)
